from pathlib import Path

BASE_DIR = Path(__file__).resolve().parent.parent

mat_state = {
    "healthy_train": {
        "start": "2020-07-31T23:52:09",
        "end": "2020-09-30T23:44:41",
        "description": "healthy state",
    },
    "healthy_test": {
        "start": "2020-09-30T23:54:40",
        "end": "2020-10-09T08:30:12",
        "description": "healthy state test period",
    },
    "damage1": {
        "start": "2020-10-13T18:29:56",
        "end": "2020-10-27T11:32:37",
        "description": "all damage mechanisms removed",
        "severity": "high",
        "location": "DAM6",
        "closest_sensor": 9,
    },
    "healthy1": {
        "start": "2020-10-27T15:46:21",
        "end": "2020-11-09T09:39:32",
        "description": "healthy state after damage",
    },
    "damage2": {
        "start": "2020-11-09T14:28:35",
        "end": "2020-11-24T09:40:41",
        "description": "all damage mechanisms removed",
        "severity": "high",
        "location": "DAM4",
        "closest_sensor": 6,
    },
    "healthy2": {
        "start": "2020-11-24T13:50:35",
        "end": "2021-03-18T08:49:45",
        "description": "healthy state after damage",
    },
    "damage3": {
        "start": "2021-03-18T15:29:36",
        "end": "2021-04-20T11:40:16",
        "description": "all damage mechanisms removed",
        "severity": "high",
        "location": "DAM3",
        "closest_sensor": 5,
    },
    "healthy3": {
        "start": "2021-04-20T16:50:09",
        "end": "2021-05-04T10:13:49",
        "description": "healthy state after damage",
    },
    "damage4": {
        "start": "2021-05-04T15:23:42",
        "end": "2021-05-19T10:45:47",
        "description": "one damage mechanism removed",
        "severity": "low",
        "location": "DAM6",
        "closest_sensor": 9,
    },
    "healthy4": {
        "start": "2021-05-19T15:55:41",
        "end": "2021-05-28T07:11:04",
        "description": "healthy state after damage",
    },
    "damage5": {
        "start": "2021-05-28T12:20:57",
        "end": "2021-06-14T06:41:59",
        "description": "one damage mechanism removed",
        "severity": "low",
        "location": "DAM4",
        "closest_sensor": 6,
    },
    "healthy5": {
        "start": "2021-06-14T12:59:03",
        "end": "2021-06-25T05:43:20",
        "description": "healthy state after damage",
    },
    "damage6": {
        "start": "2021-06-25T11:55:57",
        "end": "2021-07-12T07:16:58",
        "description": "one damage mechanism removed",
        "severity": "low",
        "location": "DAM3",
        "closest_sensor": 5,
    },
    "healthy6": {
        "start": "2021-07-12T12:26:51",
        "end": "2021-07-31T23:46:27",
        "description": "healthy state after damage",
    },
}



from __future__ import annotations

import pandas as pd


def label_events(timestamps, events):
    """
    Label each timestamp based on defined events.

    Parameters:
    - timestamps (pd.Series): Series of datetime timestamps.
    - events (dict): Dictionary of event definitions with 'start' and 'end' times.

    Returns:
    - pd.Series: Series of labels corresponding to each timestamp.
    """
    # Initialize labels with the same index as timestamps and with no event
    labels = pd.Series(index=timestamps, dtype=object)
    labels = labels.fillna("no_event")
    # Assign labels based on events
    for event_name, event_info in events.items():
        start = pd.to_datetime(event_info["start"])
        end = pd.to_datetime(event_info["end"])
        mask = (timestamps >= start) & (timestamps <= end)
        labels.loc[mask] = event_name
    return labels



from __future__ import annotations

from functools import partial

import numpy as np
import pandas as pd

from evaluation_lumo.config import mat_state
from evaluation_lumo.metrics import compute_tr, mad, median_ratio
from evaluation_lumo.utils import label_events


def prepare_dataframe(
    timestamps: pd.Series | np.ndarray,
    damage_indexs: pd.Series | np.ndarray,
    events: dict | None = None,
    train_start: str | None = None,
    train_end: str | None = None,
) -> tuple[pd.DataFrame, pd.DataFrame, dict]:
    """
    Prepare a pandas DataFrame with labeled events and split into training and full dataframes.

    Parameters:
    ----------
    timestamps : pd.Series | np.array
        Array or series of timestamps.
    damage_indexs : pd.Series | np.array
        Array or series of anomaly scores.
    events : dict | None, optional
        Dictionary with event details. Default is None. If not provided, `mat_state` from the `config` module will be used.
    train_start : str | None, optional
        Start timestamp for the training window. Default is None.
    train_end : str | None, optional
        End timestamp for the training window. Default is None.

    Returns:
    -------
    tuple[pd.DataFrame, pd.DataFrame, dict]
        - `data`: Full dataframe with labeled events.
        - `train_data`: Dataframe with training data based on the training window.
        - `events`: Updated event dictionary if not provided.
    """
    # Use mat_state if events is not provided
    if events is None:
        events = mat_state
    if train_start is None or train_end is None:
        train_start = events["healthy_train"]["start"]
        train_end = events["healthy_train"]["end"]

    # Create the dataframe and label events

    data = pd.DataFrame({"timestamp": timestamps, "score": damage_indexs})
    events_list = label_events(data["timestamp"].values, events)
    data.loc[:, "event"] = events_list.values
    train_data = data[(timestamps >= train_start) & (timestamps <= train_end)][
        "score"
    ].values

    return data, train_data


def compute_tr_by_events(
    timestamps: pd.Series | np.ndarray,
    damage_indexs: pd.Series | np.ndarray,
    fpr_train: float = 0.01,
    events: dict | None = None,  # Leave as None
    train_start: str | None = None,  # Leave as None
    train_end: str | None = None,  # Leave as None
) -> dict:
    """
    Compute the True Rate (TR) for each event in the events dictionary.
    The threshold is set to ensure the False Positive Rate (FPR) is 0.01 for healthy training data.
    """
    # Prepare the data
    data, train_data = prepare_dataframe(
        timestamps, damage_indexs, events, train_start, train_end
    )

    # Compute the threshold based on the training data
    threshold = np.quantile(train_data, 1 - fpr_train)
    # Compute True Rate for each event
    res = data.groupby("event").apply(
        lambda x: compute_tr(x["score"], threshold), include_groups=False
    )
    return res.to_dict(), threshold


def compute_median_variation(
    timestamps: pd.Series | np.ndarray,
    damage_indexs: pd.Series | np.ndarray,
    events: dict | None = None,  # Leave as None
    train_start: str | None = None,  # Leave as None
    train_end: str | None = None,  # Leave as None
) -> float:
    """
    Compute the mean variation of anomaly scores for each event in the events dictionary.
    """
    # Prepare the data
    data, train_data = prepare_dataframe(
        timestamps, damage_indexs, events, train_start, train_end
    )

    # Compute mean variation for each event
    median_ratio_partial = partial(
        median_ratio, damage_index_healthy=train_data
    )
    res = data.groupby("event").apply(
        lambda x: median_ratio_partial(damage_index_damaged=x["score"]),
        include_groups=False,
    )
    return res.to_dict()


def compute_nmad(
    timestamps: pd.Series | np.ndarray,
    damage_indexs: pd.Series | np.ndarray,
    events: dict | None = None,  # Leave as None
    train_start: str | None = None,  # Leave as None
    train_end: str | None = None,  # Leave as None
) -> dict:
    """
    Compute the  Median Absolute Deviation (MAD) for each event in the events dictionary.

    Parameters:
    ----------
    timestamps : pd.Series | np.ndarray
        Array or series of timestamps.
    damage_indexs : pd.Series | np.ndarray
        Array or series of anomaly scores.
    events : dict | None, optional
        Dictionary with event details. Default is None.
    train_start : str | None, optional
        Start timestamp for the training window. Default is None.
    train_end : str | None, optional
        End timestamp for the training window. Default is None.

    Returns:
    -------
    dict
        A dictionary with SMAD values for each event.
    """
    # Prepare the data
    data, training_data = prepare_dataframe(
        timestamps, damage_indexs, events, train_start, train_end
    )
    iqr = np.quantile(training_data, 0.99) - np.quantile(training_data, 0.01)
    # Compute SMAD for each event
    res = data.groupby("event").apply(
        lambda x: mad(x["score"]) / iqr, include_groups=False
    )

    return res.to_dict()



from __future__ import annotations

import warnings
from sys import float_info

import numpy as np
import pandas as pd

eps = float_info.epsilon


def compute_tr(damage_index: np.ndarray | pd.Series, threshold: float) -> float:
    """
    Compute the Trigger Rate (TPR/FPR) given anomaly scores and a threshold.

    Parameters:
    - damage_index (array-like): Anomaly index scores.
    - threshold (float): Threshold above which scores are considered anomalies.

    Returns:
    - float: Trigger Rate.
    """
    if isinstance(damage_index, pd.Series):
        damage_index = damage_index.values
    if isinstance(damage_index, list):
        damage_index = np.array(damage_index)

    return np.sum(damage_index > threshold) / len(damage_index)


def median_ratio(
    damage_index_healthy: np.ndarray | pd.Series,
    damage_index_damaged: np.ndarray | pd.Series,
) -> float:
    """
    Compute the mean ratio of anomaly scores between healthy and damaged states.

    Parameters:
    - damage_index_healthy (array-like): Anomaly index scores for healthy state.
    - damage_indexs_damaged (array-like): Anomaly index scores for damaged state.

    Returns:
    - float: Mean ratio of anomaly scores.
    """
    if isinstance(damage_index_healthy, pd.Series):
        damage_index_healthy = damage_index_healthy.values
    if isinstance(damage_index_damaged, pd.Series):
        damage_index_damaged = damage_index_damaged.values

    if isinstance(damage_index_healthy, list):
        damage_index_healthy = np.array(damage_index_healthy)
    if isinstance(damage_index_damaged, list):
        damage_index_damaged = np.array(damage_index_damaged)

    # Compute the range of healthy scores
    range_healthy = np.percentile(damage_index_healthy, 99) - np.percentile(
        damage_index_healthy, 1
    )

    # Handle zero range by raising a warning and using a range of 1
    if range_healthy == 0:
        warnings.warn(
            "Range of healthy anomaly scores is 0. Using a fallback range of 1.",
            UserWarning,
        )
        range_healthy = 1

    # Compute the mean ratio
    res = (
        np.median(damage_index_damaged)
        / (np.median(damage_index_healthy) + eps)
        / (range_healthy + eps)
    )
    return res


def mad(
    time_series: np.ndarray | pd.Series,
) -> float:
    """
    Compute the Median Absolute Deviation (MAD) for a time series.

    Parameters:
    - time_series (array-like): Input time series data.

    Returns:
    - float: MAD, a scale-variant measure of dispersion.
    """
    if isinstance(time_series, pd.Series):
        time_series = time_series.values
    if isinstance(time_series, list):
        time_series = np.array(time_series)

    if len(time_series) == 0:
        raise ValueError("Time series is empty.")

    median = np.median(time_series)
    mad = np.median(np.abs(time_series - median))
    return mad
